﻿
/*******************************************
 * 信号処理クラス
 * 
 * 更新履歴
 *  2010/5/23   森下整備開始
 *  2011/2/18   名前空間を名づけた
 *              今後：渡されたデータ幅がFFTデータ幅を超えた場合でも、可能なだけ結果を返すようにしたい。
 *  2011/2/26   コンストラクタでのエラー対策と、コメントを入れる。
 *              拡張予定：　帯域指定でのパワー計算値を返すメソッド（引数にサンプリング周波数情報が必須）
 *                          窓関数に三角窓も追加予定
 * ***************************************/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Signal_Process
{
    /********************************************************
     * FFT/IFFT関連クラス
     *  開発者：Katsuhiro Morishita(2011.2) Kumamoto-University
     *  
     *  参考文献:ニューメリカル・レシピ・イン・シー
     *           用語はこの書籍に準ずる。
     *  更新履歴：2010/5/23   森下整備開始
     * ******************************************************/
    /// <summary>
    /// Discrete Fourier Transformクラス     
    /// 
    /// 
    /// 使用する際には、プロジェクトにファイルを追加した後に、ソースヘッダに以下のコードを追加して下さい。
    /// 「
    /// using Signal_Process;
    /// 」
    /// </summary>
    class DFT
    {
        // 定数宣言
        /// <summary>
        ///     FFTデータ幅を指定するときに便利な列挙体
        /// </summary>
        public enum FFT_point :int{ // データサイズ(2^n)をこの列挙体を使って宣言すれば間違えない.
            size32 = 32, 
            size64 = 64,
            size128 = 128,
            size256 = 256,
            size512 = 512,
            size1024 = 1024,
            size2048 = 2048,
            size4096 = 4096,
            size8192 = 8192
        }
        private enum FFTorIFFT : int { 
            FFTconvert = 1,
            IFFTconvert = -1
        }
        public enum Window
        {
            Hanning,
            Perzen,
            NoWindow
        }
        // 構造体宣言
        public struct FFTdatas
        {
            public double Re;
            public double Im;
            public double Abs;          // 解析後に格納される
            public double Log;          // 解析後に格納される
        }
        // グローバル変数の宣言
        private int __FFTpoint;
        private Window __WindowKind;
        public FFTdatas[] calced_datas;
        private FFTdatas[] set_datas;

        /****************************　プロパティ　************************************/
        /*-----------------------------------
         * プロパティ　窓関数の種別をセット・読み取る
         * 
         * 本クラスを利用する外部側から、チェックする
         * ---------------------------------*/
        /// <summary>
        ///     窓関数の種別
        /// </summary>
        public Window WindowKind
        {
            get{
                return this.__WindowKind;
            }
            set {
                this.__WindowKind = value;
            }
        }
        /*-----------------------------------
         * プロパティ　FFT/IFFTデータ幅を読み取る
         * 
         * 本クラスを利用する外部側から、チェックする
         * ---------------------------------*/
        /// <summary>
        ///     FFT/IFFTデータ幅
        /// </summary>
        public int FFTsize
        {
            get
            {
                return this.__FFTpoint;
            }
        }
        /*-----------------------------------
         * プロパティ　変換後の配列に含まれる最大値を返す
         * 
         * 本クラスを利用する外部側から、チェックする
         * ---------------------------------*/
        /// <summary>
        ///     変換後の配列に含まれる最大値を返す
        /// </summary>
        public double Max
        {
            get
            {
                double max = 0;
                int i;

                for (i = 0; i < this.calced_datas.Length; i++) {
                    if (max < this.calced_datas[i].Abs) max = this.calced_datas[i].Abs;
                }
                return max;
            }
        }
        /*-----------------------------------
         * プロパティ　変換後の配列に含まれる平均値を返す
         * 
         * 本クラスを利用する外部側から、チェックする
         * ---------------------------------*/
        /// <summary>
        ///     変換後の配列に含まれる平均値を返す
        /// </summary>
        public double Mean
        {
            get
            {
                double mean = 0;
                int i;

                for (i = 0; i < this.calced_datas.Length; i++)
                {
                    mean += this.calced_datas[i].Abs;
                }
                return mean / this.calced_datas.Length;
            }
        }

        /****************************　メソッド　************************************/
        /********************************************
         * 窓関数　ハミング窓（Hanning window）・・・データサイズに合わせて、両端を0に落とす
         *   
         *   引数：データ配列
         *   返値：ウィンドウをかけた後のデータ配列
        * ******************************************/
        private double[] Hanning_window(double[] data, int size)
        {
            double wj;
            int j;

            for( j = 0 ; j < size ; j++ )
            {
                wj = 0.5 * (1.0 - Math.Cos(2.0 * Math.PI * (double)j / ((double)size - 1))); // 窓の計数を計算
                data[j] *= wj;
            }
            return data;
        }
        /********************************************
        * 窓関数　Parzen窓(Parzen window)三角窓・・・ハミング窓より周波数のブレが少ないが、サイドローブはやや大きい
        *   引数：データ配列
        *   返値：ウィンドウをかけた後のデータ配列
        * ******************************************/
        private double[] Parzen_window(double[] data, int size) {
            double wj;
            int j;

            for( j = 0 ; j < size ; j++ ) {
                wj = 1.0 - Math.Abs(((double)j - 0.5 * ((double)size - 1)) / (0.5 * ((double)size + 1))); // 窓の計数を計算
                data[j] *= wj;
            }
            return data;
        }
        /********************************************
         * 渡されたデータを記憶する
         *  データサイズよりFFTサイズの方が大きい場合は、0で埋める。
         *  反対に、小さい場合は入るだけしか受け取らない。
         * ******************************************/
        /// <summary>
        ///     データをセットするメソッド。
        ///     指定された窓関数をかけて保持する。
        /// </summary>
        private void Dataset(FFTdatas[] arr) {
            int i;
            int size;
            double[] temp = new double[arr.Length];

            size = arr.Length;
            if (this.__FFTpoint < arr.Length) size = this.__FFTpoint;   // フィルタにかけるサイズを決定する
            // Reパート
            for (i = 0; i < arr.Length; i++) temp[i] = arr[i].Re;
            if (this.__WindowKind == Window.Hanning)
            {
                temp = this.Hanning_window(temp, size);
            }
            else if (this.__WindowKind == Window.Perzen){
                temp = this.Parzen_window(temp, size);
            }
            for (i = 0; i < this.__FFTpoint; i++)                       // 受け渡されたデータをコピー
            {
                if( arr.Length > i ){                                   // ある限りコピー
                    this.set_datas[i].Re = temp[i];
                }else {
                    this.set_datas[i].Re = 0.0;                         // 余ったら0を詰める
                }
            }

            // Imパート
            for (i = 0; i < arr.Length; i++) temp[i] = arr[i].Im;
            if (this.__WindowKind == Window.Hanning)
            {
                temp = this.Hanning_window(temp, size);
            }
            else if (this.__WindowKind == Window.Perzen)
            {
                temp = this.Parzen_window(temp, size);
            }
            for (i = 0; i < this.__FFTpoint; i++)                       // 受け渡されたデータをコピー
            {
                if (arr.Length > i)
                {                                                       // ある限りコピー
                    this.set_datas[i].Im = temp[i];
                }
                else
                {
                    this.set_datas[i].Im = 0.0;                         // 余ったら0を詰める
                }
            }
            return;
        } 
        /********************************************
         * 渡されたデータを記憶する
         *  実部のみ(虚部のみでもいいんだけど)のデータを解析するのに使用する。
         *  また、データサイズよりFFTサイズの方が大きい場合は、0で埋める。
         * ******************************************/
        /// <summary>
        ///     データをセットするメソッド2
        ///     指定された窓関数をかけて保持する。
        /// </summary>
        private void Dataset(double[] arr) {
            int i;
            int size;
            FFTdatas zero = new FFTdatas();
            double[] temp = new double[arr.Length];

            zero.Re = 0;
            zero.Im = 0;

            size = arr.Length;
            if (this.__FFTpoint < arr.Length) size = this.__FFTpoint;   // フィルタにかけるサイズを決定する
            for (i = 0; i < arr.Length; i++) temp[i] = arr[i];
            if (this.__WindowKind == Window.Hanning)
            {
                temp = this.Hanning_window(temp, size);
            }
            else if (this.__WindowKind == Window.Perzen)
            {
                temp = this.Parzen_window(temp, size);
            }
            for (i = 0; i < this.__FFTpoint; i++)                       // 受け渡されたデータをコピー
            {
                if( arr.Length > i ){                                   // ある限りコピー
                    this.set_datas[i].Re = temp[i];
                    this.set_datas[i].Im = 0.0;
                }else {
                    this.set_datas[i] = zero;                           // 余ったら0を詰める
                }
            }
            return;
        }
        /**************************************************
         * FFT/IFFT関数
         * 引数：   data[]　非変換変数。結果も格納。長さnnの複素配列。
         *                  配列の中には、実部・虚部・実部・虚部・実部・虚部・実部・虚部・…と並んでおればよい。
         *                  原本では要素番号1から詰めていることを仮定している。
         *                  実部と虚部が入れ替わっても特に影響は無いはず。
         *          nn　    data[]に格納されたデータ長を表す。要素数の半分である。2の整数乗である必要がある。
         *          isign　 1：FFT変換を行う　-1：IFFT変換を行う。結果はnn倍される。
         ****************************************************/
        /// <summary>
        ///     FFT/IFFTメソッド
        /// </summary>
        private void Calc(double[] data, long nn , int isign ){
            long n, mmax, m, j, istep, i;
            double wtemp, wr, wpr, wpi, wi, theta;              // 三角関数の漸化式用
            double tempr, tempi;

            n = nn << 1;                                        // n:データポインタ数の2倍
            j = 1;
            for (i = 1; i < n; i += 2) {                        // ビット反転アルゴリズム
                if (j > i) {                                    // 複素数を交換
                    tempr = data[j]; data[j] = data[i]; data[i] = tempr;                    // 原本では実部を表す
                    tempi = data[j + 1]; data[j + 1] = data[i + 1]; data[i + 1] = tempi;    // 原本では虚部を表す
                }
                m = n >> 1;
                while (m >= 2 && j > m) {
                    j -= m;
                    m >>= 1;
                }
                j += m;
            }
            mmax = 2;                                           // 以下はDanielson-Lanczosアルゴリズムを採用
            while (n > mmax) {                                  // 外側のループはlog2 nn回実行される
                istep = mmax << 1;
                theta = isign * (6.28318530717959 / mmax);      // 三角関数の漸化式の初期値. 2*pi = 6.28…
                wtemp = Math.Sin (0.5 * theta);
                wpr = -2.0 * wtemp * wtemp;
                wpi = Math.Sin(theta);
                wr = 1.0;
                wi = 0.0;
                for (m = 1; m < mmax; m += 2) {
                    for (i = m; i <= n; i += istep) {
                        j = i + mmax;                           // Danielson-Lanczos公式
                        tempr = wr * data[j] - wi * data[j + 1];// 原本通りならreal part
                        tempi = wr * data[j + 1] + wi * data[j];// 原本通りならimaginary part
                        data[j] = data[i] - tempr;
                        data[j + 1] = data[i + 1] - tempi;
                        data[i] += tempr;
                        data[i+1] += tempi;
                    }                                           // 三角関数の漸化式
                    wr = (wtemp = wr) * wpr - wi * wpi + wr;    // ここは処理をひっくり返せばwtempをそのままwrに置き換えられるが、恐らくrealパートから計算したいのだろう。
                    wi = wi * wpr + wtemp * wpi + wi;
                }
                mmax = istep;
            }
            return;
        }
        /**************************************************
         * FFT/IFFT関数
         * 引数：   data[]　非変換変数。結果も格納。長さ2^nの複素配列。
         *          isign　 1：FFT変換を行う　-1：IFFT変換を行う。結果は2^n倍される。
        ****************************************************/
        /// <summary>
        ///     FFT/IFFTメソッド
        /// </summary>
        /*
        public void Calc(FFTdatas[] data, int isign)
        {
            long n, mmax, m, j, istep, i;
            double wtemp, wr, wpr, wpi, wi, theta;              // 三角関数の漸化式用
            double tempr, tempi;

            
            j = 1;
            for (i = 0; i < data.Length; i++)
            {                        // ビット反転アルゴリズム
                if (j > i)
                {                                    // 実部・虚部を交換
                    tempr = data[j].Re; data[j].Re = data[i].Re; data[i].Re = tempr;                    // 原本では実部を表す
                    tempi = data[j].Im; data[j].Im = data[i].Im; data[i].Im = tempi;    // 原本では虚部を表す
                }
                m = data.Length;
                while (m >= 2 && j > m)
                {
                    j -= m;
                    m >>= 1;
                }
                j += m;
            }
            n = data.Length << 1;                                        // n:データポインタ数の2倍
            
            mmax = 2;                                           // 以下はDanielson-Lanczosアルゴリズムを採用
            while (n > mmax)
            {                                  // 外側のループはlog2 nn回実行される
                istep = mmax << 1;
                theta = isign * (6.28318530717959 / mmax);      // 三角関数の漸化式の初期値. 2*pi = 6.28…
                wtemp = Math.Sin(0.5 * theta);
                wpr = -2.0 * wtemp * wtemp;
                wpi = Math.Sin(theta);
                wr = 1.0;
                wi = 0.0;
                for (m = 1; m < mmax; m += 2)
                {
                    for (i = m; i <= n; i += istep)
                    {
                        j = i + mmax;                           // Danielson-Lanczos公式
                        tempr = wr * data[j] - wi * data[j + 1];// 原本通りならreal part
                        tempi = wr * data[j + 1] + wi * data[j];// 原本通りならimaginary part
                        data[j] = data[i] - tempr;
                        data[j + 1] = data[i + 1] - tempi;
                        data[i] += tempr;
                        data[i + 1] += tempi;
                    }                                           // 三角関数の漸化式
                    wr = (wtemp = wr) * wpr - wi * wpi + wr;    // ここは処理をひっくり返せばwtempをそのままwrに置き換えられるが、恐らくrealパートから計算したいのだろう。
                    wi = wi * wpr + wtemp * wpi + wi;
                }
                mmax = istep;
            }
            return;
        }
         * */
        /********************************************
         * 実部のみのデータを解析する
         *  鳥の声を解析するときはこれを呼び出せばよい。
         *  例：
         *      fft a = new fft();                              // このクラスを宣言して、実態も確保
         *      a.FFT(temp);                                    // 実数配列を渡してFFT演算を実行。配列サイズがFFTデータ幅とする必要はない。
         *  
         *  FFT結果はF_reとF_imに格納される。解析後に外部からアクセスすると良い。
         * 
         *  引数： data：データ配列
         *  返値： void
         * ******************************************/
        /// <summary>
        ///     受け渡されたデータを使用してFFTを実行する
        /// </summary>
        public void  FFT(double[] data)
        {
            int i,k;
            double[] data_for_calc = new double[this.__FFTpoint * 2 + 1];

            this.Dataset(data);                                 // データをセットする
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                data_for_calc[k] = this.set_datas[i].Re;        // 解析するデータを格納する
                data_for_calc[k + 1] = 0.0;                     // こっちはimaginaryパート
            }
            this.Calc(data_for_calc, this.__FFTpoint, (int)FFTorIFFT.FFTconvert);  // FFT演算
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                this.calced_datas[i].Re = data_for_calc[k];     // 結果を格納する
                this.calced_datas[i].Im = data_for_calc[k + 1]; // こっちはimaginaryパート
                this.calced_datas[i].Abs = Math.Sqrt(Math.Pow(this.calced_datas[i].Re, 2) + Math.Pow(this.calced_datas[i].Im, 2));
                this.calced_datas[i].Log = Math.Log10(this.calced_datas[i].Abs);
            }
            return;
        }
        /// <summary>
        ///     実部・虚部を合わせて解析するFFT/IFFTメソッド
        ///     <param name="data">double型の配列に解析したいデータを格納してください</param>
        /// </summary>
        public void FFT(FFTdatas[] data) 
        {
            int i, k;
            double[] data_for_calc = new double[this.__FFTpoint * 2 + 1];

            this.Dataset(data);                                 // データをセットする
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                data_for_calc[k] = this.set_datas[i].Re;        // 解析するデータを格納する
                data_for_calc[k + 1] = this.set_datas[i].Im;    // こっちはimaginaryパート
            }
            this.Calc(data_for_calc, this.__FFTpoint, (int)FFTorIFFT.FFTconvert);  // FFT演算
            for (i = 0; i < this.__FFTpoint; i++)
            {
                k = 2 * i + 1;
                this.calced_datas[i].Re = data_for_calc[k];     // 結果を格納する
                this.calced_datas[i].Im = data_for_calc[k + 1]; // こっちはimaginaryパート
                this.calced_datas[i].Abs = Math.Sqrt(Math.Pow(this.calced_datas[i].Re, 2) + Math.Pow(this.calced_datas[i].Im, 2));
                this.calced_datas[i].Log = Math.Log10(this.calced_datas[i].Abs);
            }
            return;
        }

        /****************************　初期化/削除　************************************/
        /*****************
         * このクラスのコンストラクタ
         * 
         * 初期化時に使用します。
         * ***************/
        /// <summary>
        /// 　デフォルトでは4096ポイントFFT/IFFT，窓関数の種別としてハミングウィンドウがセットされます。
        /// </summary>
        /// <param name="FFT_point">FFT/IFFTポイント数</param>
        /// <param name="window_kind">窓関数の種類</param>
        public DFT(FFT_point FFT_point = FFT_point.size4096 , Window window_kind = Window.Hanning){
            this.__FFTpoint = (int)FFT_point;
            this.__WindowKind = window_kind;

            this.calced_datas = new FFTdatas[this.__FFTpoint];     // サイズに合わせて、インスタンスを生成
            this.set_datas = new FFTdatas[this.__FFTpoint];
        }
        /// <summary>
        /// 　デフォルトでは4096ポイントFFT/IFFT，窓関数の種別としてハミングウィンドウがセットされます。
        /// 　指定ポイント数が2^Nではない場合、FFT_point＜2^Nとなる最小の2^Nを採用しますので注意.
        /// </summary>
        /// <param name="FFT_point">FFT/IFFTポイント数</param>
        /// <param name="window_kind">窓関数の種類</param>
        public DFT(int FFT_point = 4096, Window window_kind = Window.Hanning)
        {
            int FFT_point2 = 1;

            while (FFT_point2 < FFT_point) FFT_point2 <<= 1;        // サイズ調整
            this.__FFTpoint = FFT_point2;
            this.__WindowKind = window_kind;
            
            this.calced_datas = new FFTdatas[this.__FFTpoint];      // サイズに合わせて、インスタンスを生成
            this.set_datas = new FFTdatas[this.__FFTpoint];
        }
    }
}
